local MAJOR, MINOR = "LibSimpleTimer", "1.5";
local lib = LibStub:NewLibrary(MAJOR, MINOR);

if not lib then
	return;
end

-- alarm = { duration, callback}
-- callback = function(self, time)

local TIMER_TABLE = TIMER_TABLE or {};	-- name/ID : alarm

local frame;
if not frame then
	frame = CreateFrame("Frame", "SimpleTimerFrame");
end

-------------------------------------------------------------------------------------------------------------------------SimpleTimer

local removeTab = {};
local function update(self, elapsed)
	local time = GetTime();
	local count = 0;
	for id, alarm in pairs(TIMER_TABLE) do
		count = count + 1;
		alarm.duration = alarm.duration - elapsed;
		if alarm.duration <= 0 then
			if alarm.callBack then
				alarm:callBack(time);
				tinsert(removeTab, id);
			end
		end
	end
	for _, id in ipairs(removeTab) do
		TIMER_TABLE[id] = nil;
	end
	
	if count == #removeTab then
		frame:SetScript("OnUpdate", nil);
	end
	wipe(removeTab);
end

function lib:SetDurationAlarm(id, duration, callBack)
	if not id then
		id = tostring(GetTime());
	end
	
	local alarm = {};
	alarm.duration = duration or 0;
	alarm.callBack = callBack;
	
	TIMER_TABLE[id] = alarm;
	
	if not frame:GetScript("OnUpdate") then
		frame:SetScript("OnUpdate", update);
	end
	
	return id;
end

function lib:CancelTimer(id)
	if TIMER_TABLE[id] then
		TIMER_TABLE[id].callBack = nil;
	end
end

