local function am(msg)
	ChatFrame1:AddMessage(msg);
end

local _name, _addon = ...;

local DB;
local Timer = LibStub("LibSimpleTimer");

--static vars
local FONT = "Fonts\\MORPHEUS.TTF";
local FONT_SIZE = 12;

local ANCHOR = { "TOP", "TOPRIGHT", "RIGHT", "BOTTOMRIGHT", "BOTTOM", "BOTTOMLEFT", "LEFT", "TOPLEFT",};
local REVANCHOR = { "BOTTOM", "BOTTOMLEFT", "LEFT", "TOPLEFT", "TOP", "TOPRIGHT", "RIGHT", "BOTTOMRIGHT", };

local ICON_BROKER = "Interface\\WorldStateFrame\\";
local ICON_PINGED = "Interface\\Addons\\".._name.."\\Textures\\pinged";
local ICON_UNPINGED = "Interface\\Addons\\".._name.."\\Textures\\unpinged";
local TIPBORDER = "Interface\\Addons\\".._name.."\\Textures\\border";
local TEXT_NOMESSAGE = "WorldState";

local ENTRYTABLE = {};
local ENTRYHEIGHT = 24;

local UIDELAY = 0.5;
local UIOffset_H = 8;
local UIOffset_V = 8;

local defaultDB = {
	anchor_index = 1,
	pattern = "%d+[:/]?%d*",
	separator = " \\ ";
	hideWorldStateFrame,
	pinged,
	anchorX,
	anchorY,
};

local BACKDROP = {
	bgFile = nil,--"Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = TIPBORDER,
	tileSize = 0, edgeSize = 16,
	insets = { left = 2, right = 2, top = 2, bottom = 2,	},
};

--vars
local onBlock, onTip;
local TipFrame;
-------------------------------------------------------------------------------------------------------------------------
local function getAnchoredPosition(self, point)
	point = point or self:GetPoint(1);
	local l, b, r, t = self:GetLeft(), self:GetBottom(), self:GetRight(), self:GetTop();
	local x, y;
	if string.find(point, "LEFT") then
		x = l;
	elseif string.find(point, "RIGHT") then
		x = r;
	else
		x = (l + r) / 2;
	end
	if string.find(point, "TOP") then
		y = t;
	elseif string.find(point, "BOTTOM") then
		y = b;
	else
		y = (t + b) / 2;
	end
	return point, x, y;
end

local stringBuf = {};
local subStringBuf = {};
local function updateBroker(infoTable)
	while #infoTable > 0 do
		local info = tremove(infoTable, 1);
		info:gsub(DB.pattern, function(w)
			tinsert(subStringBuf, w);
		end);
		if #subStringBuf > 0 then
			tinsert(stringBuf, table.concat(subStringBuf, " "));
			wipe(subStringBuf);
		end
	end
	if #stringBuf > 0 then
		_addon.dataobj.text = table.concat(stringBuf, DB.separator);
		wipe(stringBuf);
	else
		_addon.dataobj.text = TEXT_NOMESSAGE;
	end
end

local function updatePosition(self)
	_, DB.anchorX,	DB.anchorY = getAnchoredPosition(self, ANCHOR[DB.anchor_index]);
end

local function doShow(self)
	TipFrame:ClearAllPoints();
	if self then
		TipFrame:SetPoint(ANCHOR[DB.anchor_index], self, REVANCHOR[DB.anchor_index]);
	else
		TipFrame:SetPoint(ANCHOR[DB.anchor_index], UIParent, "BOTTOMLEFT", DB.anchorX, DB.anchorY);
	end
	TipFrame:Show();
end

local function doHide()
	Timer:CancelTimer(_name);
	if not onBlock and not onTip and not DB.pinged then
		TipFrame:Hide();
	end
end

local function startHiding()
	updatePosition(TipFrame);
	Timer:SetDurationAlarm(_name, UIDELAY, doHide);
end

local function createTipFrame()
	local f = CreateFrame("Frame", _name.."_Tip", UIParent);
	f:EnableMouse(true);
	f:SetBackdrop(BACKDROP);

	local btnTex = f:CreateTexture(nil, "ARTWORK");
	btnTex:SetTexture(DB.pinged and ICON_PINGED or ICON_UNPINGED);
	btnTex:SetWidth(ENTRYHEIGHT);
	btnTex:SetHeight(ENTRYHEIGHT);
	btnTex:SetPoint("TOPRIGHT", -2, -2);

	f:SetScript("OnMouseDown", function(self, btn)
		DB.pinged = not DB.pinged;
		btnTex:SetTexture(DB.pinged and ICON_PINGED or ICON_UNPINGED);
	end);
	f:SetScript("OnEnter", function(self)
		onTip = 1;
	end);
	f:SetScript("OnLeave", function(self)
		onTip = nil;
		startHiding();
	end);

	return f;
end

local entryPool = {};
local function getEntrey(parent)
	local entry = tremove(entryPool);
	if entry then
		return entry;
	end

	entry = CreateFrame("Frame", _name..GetTime(), parent);
	entry:SetHeight(ENTRYHEIGHT);

	local icon = entry:CreateTexture(nil, "ARTWORK");
	icon:SetTexture(nil);
	icon:SetWidth(ENTRYHEIGHT);
	icon:SetHeight(ENTRYHEIGHT);
	icon:SetPoint("LEFT", UIOffset_H, 0);
	entry.icon = icon;

	local text = entry:CreateFontString(nil, "OVERLAY");--, "GameFontNormalSmall");
	text:SetFont(FONT, FONT_SIZE);
	text:SetJustifyH("LEFT");
	text:SetJustifyV("LEFT", 0, 0);
	text:SetPoint("LEFT", icon, "RIGHT", 0, 0);
	text:SetText("empty");
	entry.text = text;

	local dynamicIcon = entry:CreateTexture(nil, "ARTWORK");
	dynamicIcon:SetTexture(nil);
	dynamicIcon:SetWidth(ENTRYHEIGHT);
	dynamicIcon:SetHeight(ENTRYHEIGHT);
	dynamicIcon:SetPoint("RIGHT", -UIOffset_H, 0);
	entry.dynamicIcon = dynamicIcon;

	local dynamicIconFlash = entry:CreateTexture("$ParentFlash", "ARTWORK");
	dynamicIconFlash:SetTexture(nil);
	dynamicIconFlash:SetWidth(ENTRYHEIGHT);
	dynamicIconFlash:SetHeight(ENTRYHEIGHT);
	dynamicIconFlash:SetPoint("RIGHT", -UIOffset_H, 0);
	dynamicIconFlash:SetBlendMode("ADD");
	entry.dynamicIconFlash = dynamicIconFlash;

	return entry;
end

local info = {};
local function mirrorUpdate()
	--release
	for _, entry in ipairs(ENTRYTABLE) do
		entry:ClearAllPoints();
		entry:Hide();
		tinsert(entryPool, entry);
	end
	wipe(ENTRYTABLE);
	--rebuild
	local stateCount = 0;
	local width = TipFrame:GetWidth() - 2 * UIOffset_H;

	for i=1, GetNumWorldStateUI() do
		local uiType, state, hidden, text, icon, dynamicIcon, tooltip, dynamicTooltip, extendedUI, extendedUIState1, extendedUIState2, extendedUIState3 = GetWorldStateUIInfo(i);
		if ( (uiType ~= 1) or ((WORLD_PVP_OBJECTIVES_DISPLAY == "1") or (WORLD_PVP_OBJECTIVES_DISPLAY == "2" and IsSubZonePVPPOI()) or (instanceType == "pvp")) ) then
			if state > 0 and extendedUI == "" then
				stateCount = stateCount + 1;
				local entry = getEntrey(TipFrame);
				entry:SetPoint("TOP", 0, (1 - stateCount) * ENTRYHEIGHT - UIOffset_V);
				entry:SetPoint("LEFT");
				entry:SetPoint("RIGHT");

				entry.icon:SetTexture(icon);
				entry.icon:SetTexCoord(0, 0.6, 0, 0.6);--(0, 0.6875, 0, 0.6875);
				entry.text:SetText(text);
				tinsert(info, text);
				entry.dynamicIcon:SetTexture(dynamicIcon);
				entry.dynamicIconFlash:SetTexture(dynamicIcon and (dynamicIcon.."Flash") or nil);
				if ( state == 2 ) then
					UIFrameFlash(entry.dynamicIconFlash, 0.5, 0.5, -1);
					entry.dynamicIcon:Show();
					entry.dynamicIconFlash:Show();
				elseif ( state == 3 ) then
					UIFrameFlashStop(entry.dynamicIconFlash);
					entry.dynamicIcon:Show();
					entry.dynamicIconFlash:Show();
				else
					UIFrameFlashStop(entry.dynamicIconFlash);
					entry.dynamicIcon:Hide();
					entry.dynamicIconFlash:Hide();
				end

				entry:Show();
				tinsert(ENTRYTABLE, entry);
				width = max(width, entry.icon:GetWidth() + entry.text:GetStringWidth() + entry.dynamicIcon:GetWidth());
			end
		end
	end

	if stateCount > 0 then
		TipFrame:SetWidth(width + 2 * UIOffset_H);
		TipFrame:SetHeight(stateCount * ENTRYHEIGHT + 2 * UIOffset_V);
		TipFrame:SetAlpha(1);
	else
		TipFrame:SetWidth(2 * UIOffset_H);
		TipFrame:SetHeight(2 * UIOffset_V);
		TipFrame:SetAlpha(0);
	end

	updateBroker(info);
	wipe(info);
end

local function hookWorldStateAlwaysUpFrame()
	hooksecurefunc("WorldStateAlwaysUpFrame_Update", mirrorUpdate);
end

-------------------------------------------------------------------------------------------------------------------------
local addon = CreateFrame("Frame", nil, UIParent);
addon:UnregisterAllEvents();
addon:Hide();

function addon:ADDON_LOADED(name)
	if name ~= _name then
		return;
	end
	self:UnregisterEvent("ADDON_LOADED");

	--check existed table
	if not Broker_WorldStateDB then
		Broker_WorldStateDB = CopyTable(defaultDB);
	end
	DB = Broker_WorldStateDB;

	self:RegisterEvent("PLAYER_LOGIN");
end

function addon:PLAYER_LOGIN()
	self:UnregisterEvent("PLAYER_LOGIN");

	local factionGroup = UnitFactionGroup("player");
	ICON_BROKER = ICON_BROKER..factionGroup.."Icon";

	TipFrame = createTipFrame();

	--create LDB object
	_addon.dataobj = LibStub("LibDataBroker-1.1"):NewDataObject(_name, {
		type = "data source",
		label = _name,
		text = TEXT_NOMESSAGE,
		icon = ICON_BROKER,
		OnClick = function(self, button)
			if button == "LeftButton" then
				DB.hideWorldStateFrame = not DB.hideWorldStateFrame;
				if DB.hideWorldStateFrame then
					WorldStateAlwaysUpFrame:Hide();
				else
					WorldStateAlwaysUpFrame:Show();
				end
			elseif button == "RightButton" then
				DB.anchor_index = (DB.anchor_index + 1) % #ANCHOR;
				DB.anchor_index = DB.anchor_index == 0 and #ANCHOR or DB.anchor_index;
				doShow(self);
			end
		end,
		OnEnter = function(self)
			onBlock = 1;
			doShow(self);
		end,
		OnLeave = function(self)
			onBlock = nil;
			startHiding();
		end,
	});

	hookWorldStateAlwaysUpFrame(TipFrame);
	if DB.hideWorldStateFrame then
		WorldStateAlwaysUpFrame:Hide();
	end
	if DB.pinged then
		doShow();
	end
end

addon:SetScript("OnEvent", function(self, event, ...)
	local f = self[event];
	if f then
		f(self,...);
	end
end);

addon:RegisterEvent("ADDON_LOADED");

-------------------------------------------------------------------------------------------------------------------------
local Pattern_Arg = "[\'\"]([^\"\']*)[\'\"]";

SLASH_WORLDSTATE1 = "/ws";
SlashCmdList.WORLDSTATE = function(msg, editBox)
	if msg and strlen(msg) > 0 then
		local _, ep = msg:find("-p");
		local _, es = msg:find("-s");
		if ep then
			DB.pattern = string.match(msg, Pattern_Arg, ep) or "";
			DB.pattern = strlen(DB.pattern) > 0 and DB.pattern or defaultDB.pattern;
			am("Broker_WorldState : change pattern to \'"..DB.pattern.."\'");
		end
		if es then
			DB.separator = string.match(msg, Pattern_Arg, es) or "";
			DB.separator = strlen(DB.separator) > 0 and DB.separator or defaultDB.separator;
			am("Broker_WorldState : change separator to \'"..DB.separator.."\'");
		end
		if ep or es then
			return mirrorUpdate();
		end
	end
	am("Broker_WorldState : /ws [-p | \"new pattern\"] [-s | \"new separator\"]");
end